;; Generated package description from xeft.el  -*- no-byte-compile: t -*-
(define-package "xeft" "3.3.0.20230220.134753" "Deft feat. Xapian" '((emacs "26.0")) :commit "2655af67d2a52e2944cededab03bfd8da1a8b58c" :authors '(("Yuan Fu" . "casouri@gmail.com")) :maintainer '("Yuan Fu" . "casouri@gmail.com") :keywords '("applications" "note" "searching") :url "https://sr.ht/~casouri/xeft")
